/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/motionest_log.hh"
#include <math.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include "global.hh"

struct MEData_LogSearch : public MEData
{
  ~MEData_LogSearch() { delete bcdata; }

  BCData*     bcdata;
  SearchParam sparam;
};


MotionEstimator_Logarithmic::MotionEstimator_Logarithmic()
{
  d_blkcmp = BlockComparator::CreateDefaultBC();
}

MotionEstimator_Logarithmic::~MotionEstimator_Logarithmic()
{
  if (d_blkcmp) delete d_blkcmp;
}


MEData* MotionEstimator_Logarithmic::CreateMEData(const SearchParam& spar,const MEData*)
{
  MEData_LogSearch* medata = new MEData_LogSearch;
  medata->bcdata = d_blkcmp->CreateBCData(spar);
  medata->sparam = spar;

  return medata;
}

void MotionEstimator_Logarithmic::SetSearchParams(MEData* medata, const Bitmap<Pixel>* bitmap,
                                                  bool frame, bool top)
{
  MEData_LogSearch* med = dynamic_cast<MEData_LogSearch*>(medata);
  assert(med);

  c_medata = med;
  c_bcdata = med->bcdata;
  c_frame_flag = frame;
  c_top_flag   = top;
  c_img_width  = bitmap->AskWidth();
  c_img_height = bitmap->AskHeight();

  c_pixels_per_block = med->sparam.blk_width * med->sparam.blk_height;

  d_blkcmp->SetImageBitmap(*bitmap,frame,top);
}

bool MotionEstimator_Logarithmic::Estimate(int x,int y, int& mx,int& my,
					   double& error,double upper_error_bound)
{
  // Suchbereich ggf. verkleinern, wenn wir uns zu nahe am Rand befinden.

  int left,right,top,bottom;
  AdjustRange(c_medata->sparam, x,y, c_img_width, c_img_height, left,right,top,bottom);


  // Position und Groesse des Blocks im Originalbild festlegen.

  BlockComparator::ComparePos cmppos;
  cmppos.img_x = x;
  cmppos.img_y = y;

  // maximal erlaubter Gesamtfehler
  uint32 max_mv_error = (uint32)(upper_error_bound * c_pixels_per_block);
  // cout << "\n:max: " << max_mv_error << endl;

  uint32 nomv_error = 0;  // Fehler, falls der alte (uebergebene) Motionvektor benutzt wird.
  bool   nomv_vector_valid = false;   // Ob, der alte Motionvektor benutzt werden darf.

  int halfpelshiftfact = (c_medata->sparam.halfpel ? 1 : 0);


  // MAE-Fehler berechnen, fuer den Fall, dass der MV nicht veraendert wird.

  if (mx>=left && mx<=right && my>=top && my<=bottom)
    {
      cmppos.ref_x = (x<<halfpelshiftfact)+mx;
      cmppos.ref_y = (y<<halfpelshiftfact)+my;

      // Fehler mit altem MV berechnen.

      if (d_blkcmp->Error(c_bcdata,cmppos,nomv_error,max_mv_error))
	{
          //cout << "\n>max> " << max_mv_error << endl;
	  double mean_nomv_error = ((double)nomv_error)/c_pixels_per_block;

          //cout << "Mean_0DMV_Error: " << mean_nomv_error << endl;

	  /* Wenn der Fehler sehr gering ist, dann gleich diesen Vektor nehmen und nicht
	     weitersuchen. */
	  if (mean_nomv_error <= 4.0)  // Empfohlener Wert: 1
	    {
	      return true;
	    }
	  else // sonst die Schwelle fuer den zu suchenden MV herabsetzen.
	    if (mean_nomv_error <= 3.0)  // Empfohlener Wert: 3
	      {
		max_mv_error = nomv_error/2;
                //cout << "max1: " << max_mv_error << endl;
	      }
	    else
	      {
		max_mv_error = nomv_error*29/32;  // *0.9
                //cout << "max2: " << max_mv_error << endl;
	      }
	  
	  nomv_vector_valid = true;
	}
      else
	{ /*cout << "HOHER FEHLER\n";*/ }
    }

  // MVs durchsuchen (logarithmische Suche)

  int hstep = max(-left,right )/2;   // aktuelle horizontale Schrittweite
  int vstep = max(-top ,bottom)/2;   // aktuelle vertikale   Schrittweite
  int motx=0,moty=0;  // aktuelles Zentrum der Suche
  bool first=true;    // Nur fuer in erstem Schritt der MV-Suche ==true;

  uint32 err[3][3];   // Fehler der 9 Suchpositionen. Zentrum liegt in [1][1].

  int x0 = (x<<halfpelshiftfact); // Zentrumskoordinaten (im Referenzbild)
  int y0 = (y<<halfpelshiftfact);

  int    minx=1,miny=1; // Beste Suchposition, minx,miny aus 0,1,2
                        // (Initialisierung nicht notwendig, beruhigt aber den Compiler)
  uint32 minval=MASK31; // Fehler bei der besten Suchposition.

  // Schrittweite der Log.-Suche bis Schrittweite 1 verfeinern.
  while (hstep>0 || vstep>0)
    {
      /*
	int subsampling=1;
	if (hstep>=8) subsampling=4;
	else if (hstep>=4) subsampling=2;
	else               subsampling=1;
      */

      // 8 Positionen um das Zentrum (+ Zentrum beim ersten Mal) testen.
      for (int dx=-1;dx<=1;dx++)
	for (int dy=-1;dy<=1;dy++)
	  {
	    if (!first && dx==0 && dy==0) continue;  // Zentrum nur beim ersten mal testen.

	    int sx = motx + dx*hstep;     // zu testender Motion-Vektor ist (sx;sy)
	    int sy = moty + dy*vstep;

	    // Nur testen, wenn MV innerhalb der erlaubten Grenzen.
	    if (sx>=left && sx<=right &&
		sy>=top  && sy<=bottom)
	      {
		// Position des Blocks im Referenzbild
		cmppos.ref_x = sx + x0;
		cmppos.ref_y = sy + y0;

		// Fehler berechnen
		if (d_blkcmp->Error(c_bcdata,cmppos,err[dx+1][dy+1],minval /*,subsampling*/))
		  if (err[dx+1][dy+1]<minval)
		    {
		      // Fehler ist besser als alles bisherige -> Position und Fehler merken
		      minx=dx+1; miny=dy+1;
		      minval=err[dx+1][dy+1];
		    }
	      }
	  }

      // Beste Position fuer naechsten Suchschritt ins Zentrum verschieben

      // Zentrum der Suche verschieben
      motx += (minx-1)*hstep;
      moty += (miny-1)*vstep;

      err[1][1] = err[minx][miny]; // Fehler ist ja schon bekannt.
      minx=1;
      miny=1;

      // Suchschrittweite halbieren
      hstep /= 2;
      vstep /= 2;
    }

  /* Falls neuer Vektor nicht oder nur unwesentlich (s.o.) besser ist, als alter MV,
     den alten MV weiterbenutzen. */
  if (nomv_vector_valid && err[1][1] > max_mv_error)
    {
      error = ((double)nomv_error)/c_pixels_per_block;
      return true;
    }

  /* Falls aber der neue Vektor besser ist, den neuen verwenden. */
  if (err[1][1] <= max_mv_error)
    {
      // MV zurueckgeben
      mx = motx;
      my = moty;

      error = ((double)err[1][1])/c_pixels_per_block;
      return true;
    }

  cout << upper_error_bound << endl;
  cout << err[1][1] << " > " << max_mv_error << endl;

  // Sonst konnte leider kein MV mit der angegebenen oberen Fahlerschranke gefunden werden.
  return false;
}
