/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "coding/taskmbtype.hh"
#include "coding/mvutils.hh"
#include "control/datapool.hh"
#include "mpeg/constants.hh"
#include "options.hh"


void Task_MBType_TotalPrediction::Execute(ProcessData& data)
{
  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);

  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("MBType (total pred)",data.threadnr);
      cerr << "frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  for (int y=0;y<img->mblks.AskHeight();y++)
    for (int x=0;x<img->mblks.AskWidth();x++)
      {
        Macroblock& mb = img->mblks.Ask(y,x);

        if (img->pichdr.m_picture_coding_type == PICTYPE_I)
          mb.m_IsIntra=true;

        if (img->pichdr.m_picture_coding_type == PICTYPE_P)
          {
            mb.m_IsIntra=false;
            mb.m_HasMotionForward=true;
          }

        if (img->pichdr.m_picture_coding_type == PICTYPE_B)
          {
            mb.m_IsIntra=false;
            mb.m_HasMotionForward=true;
            mb.m_HasMotionBackward=true;
          }
      }
}



template <class Pel> uint32 MBVariance(const Pel*const* pix,int x0,int y0)
{
  uint32 var=0,mean=0;

  for (int y=0;y<16;y++)
    for (int x=0;x<16;x++)
      {
        Pel p = pix[y+y0][x+x0];
        mean += p;
        var  += p*p;
      }

  var = var - mean*mean/256;

  return var;
}

template uint32 MBVariance(const DiffPixel*const* pix,int x0,int y0);
template uint32 MBVariance(const Pixel*const* pix,int x0,int y0);




void Task_MBType_Variance::Execute(ProcessData& data)
{
  ImageData* img = ImageData::GetNr(m_framenr);
  assert(img);

  if (options.ProgressDisplay >= Options::ProgressDisplay_Normal)
    {
      ShowTaskInformation("MBType (variance)",data.threadnr);
      cerr << "frame " << m_framenr << endl;
      EndShowTaskInformation();
    }

  const DiffPixel*const* pix_forward       = NULL;
  const DiffPixel*const* pix_backward      = NULL;
  const DiffPixel*const* pix_bidirectional = NULL;

  if (img->pichdr.m_picture_coding_type == PICTYPE_P ||
      img->pichdr.m_picture_coding_type == PICTYPE_B)
    {
      pix_forward = img->pred_forw.AskBitmap_const(Bitmap_Y).AskFrame_const();
    }

  if (img->pichdr.m_picture_coding_type == PICTYPE_B)
    {
      pix_backward      = img->pred_backw.AskBitmap_const(Bitmap_Y).AskFrame_const();
      pix_bidirectional = img->pred_bidir.AskBitmap_const(Bitmap_Y).AskFrame_const();
    }


  int i_cnt = 0;
  int f_cnt = 0;
  int b_cnt = 0;
  int d_cnt = 0;

  for (int y=0;y<img->mblks.AskHeight();y++)
    for (int x=0;x<img->mblks.AskWidth();x++)
      {
        Macroblock& mb = img->mblks.Ask(y,x);

        enum { Intra, Forward, Backward, Bidirectional } PredictionType = Intra;

        if (img->pichdr.m_picture_coding_type == PICTYPE_P)
          {
            assert(pix_forward);
            uint32 intra_variance   = img->variance.Ask(y,x);
            uint32 forward_variance = MBVariance(pix_forward,x*16,y*16);

            if (forward_variance < intra_variance)
              PredictionType = Forward;
          }

        if (img->pichdr.m_picture_coding_type == PICTYPE_B)
          {
            assert(pix_forward);
            uint32 intra_variance    = img->variance.Ask(y,x);
            uint32 forward_variance  = MBVariance(pix_forward      ,x*16,y*16);
            uint32 backward_variance = MBVariance(pix_backward     ,x*16,y*16);
            uint32 bidir_variance    = MBVariance(pix_bidirectional,x*16,y*16);

            uint32 min_variance = min(min(intra_variance   ,forward_variance),
                                      min(backward_variance,bidir_variance));

            if (min_variance == intra_variance)
              PredictionType = Intra;
            else
            if (min_variance == forward_variance)
              PredictionType = Forward;
            else
            if (min_variance == backward_variance)
              PredictionType = Backward;
            else
            if (min_variance == bidir_variance)
              PredictionType = Bidirectional;
          }


        switch (PredictionType)
          {
          case Intra:
            mb.m_IsIntra = true;
            mb.m_HasMotionForward  = false;
            mb.m_HasMotionBackward = false;
            mb.m_FieldDCT = false;
            mb.SetFrameMotionType(Macroblock::FRMT_FrameBased);
            i_cnt++;
            break;
          case Forward:
            mb.m_IsIntra = false;
            mb.m_HasMotionForward  = true;
            mb.m_HasMotionBackward = false;
            mb.m_FieldDCT = false;
            mb.SetFrameMotionType(Macroblock::FRMT_FrameBased);
            f_cnt++;
            break;
          case Backward:
            mb.m_IsIntra = false;
            mb.m_HasMotionForward  = false;
            mb.m_HasMotionBackward = true;
            mb.m_FieldDCT = false;
            mb.SetFrameMotionType(Macroblock::FRMT_FrameBased);
            b_cnt++;
            break;
          case Bidirectional:
            mb.m_IsIntra = false;
            mb.m_HasMotionForward  = true;
            mb.m_HasMotionBackward = true;
            mb.m_FieldDCT = false;
            mb.SetFrameMotionType(Macroblock::FRMT_FrameBased);
            d_cnt++;
            break;
          }
      }

  if (options.ProgressDisplay >= Options::ProgressDisplay_Maximal)
    {
      cout << i_cnt << " Intra, "
           << f_cnt << " Forward, "
           << b_cnt << " Backward, "
           << d_cnt << " Bidirectional\n";
    }

  ReduceMVRange(*img);
}

