/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "control/datapool.hh"
#include "coding/motionest_log.hh"
#include "coding/motionest_ots.hh"
#include "coding/motionest_full.hh"
#include "coding/motionest_1dfs.hh"
#include "coding/motionest_ntss.hh"
#include "coding/motionest_0.hh"
#include "options.hh"
#include <assert.h>

uint32 ImageData::s_nextnr = 0;
Queue<ImageData*> ImageData::s_imgdata;
Mutex ImageData::s_poolaccess_mutex;


ImageData::ImageData()
  : nr(s_nextnr),
    no_forward_prediction(false),
    scenechange_qscale_factor(1.0),
    medata(NULL),
    prev_ip(NULL),
    next_ip(NULL),
    d_start_gop(false),
    d_is_closed_gop(false),
    d_motionestim(NULL)
{
  ImageData* me = this;
  s_imgdata.Append(me);
  s_nextnr++;
}

ImageData::~ImageData()
{
  if (source) delete source;
  if (d_motionestim) delete d_motionestim;
  if (medata) delete medata;
}

ImageData* ImageData::GetNr(uint32 nr)
{
  if (nr>=s_nextnr)
    return NULL;

  if (nr<s_imgdata.AskHead()->nr)
    return NULL;

  s_poolaccess_mutex.Lock();
  ImageData* imgdata = s_imgdata.AskFromHead(GetIndexOfFrame(nr));
  s_poolaccess_mutex.Unlock();

  return imgdata;
}

void ImageData::DeleteFrame(uint32 nr)
{
  s_poolaccess_mutex.Lock();

  int idx = GetIndexOfFrame(nr);
  delete s_imgdata.AskFromHead(idx);
  s_imgdata.RemoveFromHead(idx);

  s_poolaccess_mutex.Unlock();
}


int ImageData::GetIndexOfFrame(uint32 framenr)
{
  // TODO: Koennte man auch durch binaere Suche machen

  for (uint32 i=0;i<s_imgdata.Length();i++)
    {
      if (s_imgdata.AskFromHead(i)->nr==framenr)
        { return i; }
    }

  cerr << "ILLEGAL: " << framenr << endl;
  assert(0); // Die 'framenr' gibt es nicht in der Queue.
}

void ImageData::GetGOPHdr(GOPHeaderData& gophdr) const
{
  gophdr.m_closed_gop  = IsClosedGOP();
  gophdr.m_broken_link = false;
}


class MotionEstimator* ImageData::GetMotionEstimator()
{
  d_motionest_access_mutex.Lock();

  if (d_motionestim)
    {
      MotionEstimator* est = d_motionestim;
      d_motionestim=NULL;
      d_motionest_access_mutex.Unlock();

      return est;
    }
  else
    {
      d_motionest_access_mutex.Unlock();

      switch (options.MVSearchAlgo)
        {
        case Options::TSS:
          return new MotionEstimator_Logarithmic;
        case Options::NTSS:
          return new MotionEstimator_NTSS;
        case Options::OTS:
          return new MotionEstimator_OTS;
        case Options::ODFS:
          return new MotionEstimator_1DFS;
        case Options::FullSearch:
          return new MotionEstimator_Full;
        case Options::NullMV:
          return new MotionEstimator_0;
        default:
          assert(0);
          break;
        }
    }
}


void ImageData::GiveBackMotionEstimator(class MotionEstimator* estim)
{
  d_motionest_access_mutex.Lock();

  if (d_motionestim==NULL)
    {
      d_motionestim=estim;
      d_motionest_access_mutex.Unlock();
    }
  else
    {
      d_motionest_access_mutex.Unlock();
      delete estim;
    }
}


void SeqHdrPool::SetSeqHdr(int framenr,const SequenceHeaderData& hdr)
{
  if (!seqhdrpos_queue.IsEmpty())
    {
      // TODO: SeqHdr an richtige Position in Queue einsortieren
      assert(seqhdrpos_queue.AskTail() < framenr);
    }

  SequenceHeaderData* shdr = new SequenceHeaderData;
  *shdr = hdr;
  seqhdr_queue.Append(shdr);
  seqhdrpos_queue.Append(framenr);
}

bool SeqHdrPool::IsSeqHdrBeforeFrame(int framenr)
{
  for (uint32 i=0;i<seqhdrpos_queue.Length();i++)
    if (seqhdrpos_queue.AskFromHead(i)==framenr)
      return true;
  return false;
}

void SeqHdrPool::ForgetSeqHdrBefore(int framenr)
{
  for (uint32 i=0;i<seqhdrpos_queue.Length();i++)
    if (seqhdrpos_queue.AskFromHead(i)==framenr)
      {
        seqhdr_queue.RemoveFromHead(i);
        seqhdrpos_queue.RemoveFromHead(i);
        return;
      }

  assert(0);
}

void SeqHdrPool::AskSeqHdr(int framenr,SequenceHeaderData& hdr)
{
  for (int i=seqhdrpos_queue.Length()-1;i>=0;i--)
    if (seqhdrpos_queue.AskFromHead(i)<=framenr)
      {
        hdr = *seqhdr_queue.AskFromHead(i);
        return;
      }

  assert(0);
}

Queue<SequenceHeaderData*> SeqHdrPool::seqhdr_queue;
Queue<int> SeqHdrPool::seqhdrpos_queue;




template class Array2<uint16>;
template class Array2<uint32>;
//template class Array2<int32>;
template class Queue<ImageData*>;

#include "image/array2.cc"
#include "utility/queue.cc"
