/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/


#include "input/preproc_v4l.hh"
#include "image/image.hh"

#include "arch/MMX/mmx.hh"
#include "arch/MMX/mmxlib.icc"

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <time.h>

/************* 1) HIER DEN PFAD GGF AENDERN !!! **********/
/************* 2) Dieses Include muss unbedingt am Schluss stehen, sonst gibt es
                  haufenweise Typ-Konflikte mit vielen Typen aus verschiedenen Includes. */
#include <linux/videodev.h>



InputPreprocessor_V4L::V4LParameter::V4LParameter()
  : m_width(12*16), m_height(9*16),
    m_nFrames(20), m_framerate100(2500),
    m_decimation_rate(2)
{
}

InputPreprocessor_V4L::InputPreprocessor_V4L(const V4LParameter& p)
  : d_parameter(p),
    d_next_num(0),
    d_thread_running(false)
{
  sprintf(d_next_name,"V4L-frame%05d",0);
}


InputPreprocessor_V4L::~InputPreprocessor_V4L()
{
}


bool InputPreprocessor_V4L::IsProgressiveSequence() const
{
  return d_parameter.m_height < 400;
}


bool InputPreprocessor_V4L::HasMoreInput() const
{
  if (d_next_num >= d_parameter.m_nFrames)
    { return false; }
  else
    { return true; }
}

class Image<Pixel>* InputPreprocessor_V4L::GetNextImage(class Image<Pixel>* reuse)
{
  if (!d_thread_running)
    { d_thread_running=true; d_grabthread.Start(StartGrabThread,this); }

  for (;;)
    {
      d_queueaccess.Lock();
      if (d_imagequeue.IsEmpty())
        {
          d_queueaccess.Unlock();

          struct timespec ts;
          ts.tv_sec=0;
          ts.tv_nsec=50000000; // 50ms
          nanosleep(&ts,NULL);
        }
      else
        {
          Image<Pixel>* img = d_imagequeue.AskHead();
          d_imagequeue.RemoveHead();

          d_queueaccess.Unlock();

          d_next_num++;
          sprintf(d_next_name,"V4L-frame%05d",d_next_num);

          if (!HasMoreInput())
            {
              d_grabthread.Stop();
              d_grabthread.Join();

              while (!d_imagequeue.IsEmpty())
                {
                  delete d_imagequeue.AskHead();
                  d_imagequeue.RemoveHead();
                }
            }

          return img;
        }
    }
}


void* InputPreprocessor_V4L::StartGrabThread(void* arg)
{
  ((InputPreprocessor_V4L*)(arg))->GrabThread();
  return NULL;
}

void InputPreprocessor_V4L::GrabThread()
{
  int nFramesGrabbed=0;

  int fd;
  unsigned char* map;
  struct video_mbuf vidmbuf;


  fd = open("/dev/video",O_RDWR);
  if (fd==-1)
    { perror("open (device): "); exit(10); }

  if (-1 == ioctl(fd,VIDIOCGMBUF,&vidmbuf)) {
    perror("ioctl VIDIOCGMBUF");
  }

  map = mmap(0,vidmbuf.size, PROT_READ|PROT_WRITE,MAP_SHARED,fd,0);
  if ((unsigned char*)-1 == map) {
    perror("mmap");
  }



  // Grabbing starten

  struct video_mmap vidmmap;
  vidmmap.width =AskWidth();
  vidmmap.height=AskHeight();
  vidmmap.format = VIDEO_PALETTE_YUV422;
      
  for (int i=0;i<vidmbuf.frames;i++)
    {
      vidmmap.frame =i;
      if (-1 == ioctl(fd,VIDIOCMCAPTURE,&vidmmap)) {
        perror("ioctl VIDIOCMCAPTURE");
        exit(10);
      }
    }



  int decimationcnt=0;
  int bufferidx = 0;

  while (nFramesGrabbed < d_parameter.m_nFrames && !d_grabthread.ShallIStop())
    {
      vidmmap.frame=bufferidx;
      if (-1 == ioctl(fd,VIDIOCSYNC,&vidmmap)) {
        perror("ioctl VIDIOCSYNC");
        exit(10);
      }

      if (decimationcnt==0)
        {
          if (d_imagequeue.Length()>=100 /*DefaultValue*/)
            {
              cerr << "WARNING: Image-queue overflow\n";
            }
          else
            {
              Image<Pixel>* img = new Image<Pixel>;
              img->CreateImage(YCrCb , AskWidth(),AskHeight(),true,true);
              
              Bitmap<Pixel>& yy = img->AskBitmap(Bitmap_Y);
              Bitmap<Pixel>& cb = img->AskBitmap(Bitmap_Cb);
              Bitmap<Pixel>& cr = img->AskBitmap(Bitmap_Cr);
              
              unsigned char* m=map + vidmbuf.offsets[bufferidx];
              int h_half = AskHeight()/2;
              int w_half = AskWidth() /2;

              Pixel** ypp  = yy.AskFrame();
              Pixel** cbpp = cb.AskFrame();
              Pixel** crpp = cr.AskFrame();

              if (MMX_Supported)
                {
                  SetMM0_00FFs();

                  for (int y=0;y<h_half;y++)
                    {
                      Pixel* yp  = ypp [2*y];
                      Pixel* cbp = cbpp[y];
                      Pixel* crp = crpp[y];
                      
                      for (int x=0;x<w_half;x++)
                        {
                          yp[2*x]   = m[0];
                          yp[2*x+1] = m[2];
                          cbp[x]    = m[1];
                          crp[x]    = m[3];
                          
                          m+=4;
                        }
                      
                      yp  = ypp[2*y+1];
                      
                      for (int x=0;x<w_half;x+=4)
                        {
                          __asm__
                            (
                             "movq  (%1),%%mm1\n\t"
                             "movq 8(%1),%%mm2\n\t"
                             "pand %%mm0,%%mm1\n\t"
                             " pand %%mm0,%%mm2\n\t"
                             "packuswb %%mm2,%%mm1\n\t"
                             "movq %%mm1,(%0)\n\t"

                             : : "r" (yp), "r"(m)
                             );

                          yp+=8;
                          m+=16;
                        }
                    }

                  EMMS();
                }
              else
                {
                  for (int y=0;y<h_half;y++)
                    {
                      Pixel* yp  = ypp [2*y];
                      Pixel* cbp = cbpp[y];
                      Pixel* crp = crpp[y];
                      
                      for (int x=0;x<w_half;x++)
                        {
                          yp[2*x]   = m[0];
                          yp[2*x+1] = m[2];
                          cbp[x]    = m[1];
                          crp[x]    = m[3];
                          
                          m+=4;
                        }
                      
                      yp  = ypp[2*y+1];
                      
                      for (int x=0;x<w_half;x++)
                        {
                          yp[2*x]   = m[0];
                          yp[2*x+1] = m[2];
                          
                          m+=4;
                        }
                    }
                }
              
              img->IsFramePicture = true;
              
              d_queueaccess.Lock();
              d_imagequeue.Append(img);
              d_queueaccess.Unlock();
              
              decimationcnt = d_parameter.m_decimation_rate-1;
              nFramesGrabbed++;
            }
        }
      else
        {
          decimationcnt--;
        }

      vidmmap.frame = bufferidx;
      if (-1 == ioctl(fd,VIDIOCMCAPTURE,&vidmmap)) {
        perror("ioctl VIDIOCMCAPTURE");
        exit(10);
      }

      bufferidx++;
      if (bufferidx==vidmbuf.frames)
        bufferidx=0;
    }

  munmap((char*)map,vidmbuf.size);
  close(fd);

  cout << "Grabbing stopped.\n";

  d_grabthread.ImStopping();
}


template class Queue<Image<Pixel>*>;

#include "utility/queue.cc"
