/********************************************************************************
    Copyright (C) 1998/1999 Dirk Farin

    This program is distributed under GNU Public License (GPL) as
    outlined in the COPYING file that comes with the source distribution.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ********************************************************************************/

#include "input/remote.hh"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <stdio.h>
#include <assert.h>
#include <time.h>

static double bitlength=(1.0/(38400.0/10*8)) * 8.0 / 3.0;

static struct
{
  enum IRCode code;
  const char* seq_start;
  const char* seq_step;
  const char* seq_end;
  int waittime;  // Wartezeit fuer Bildstabilisierung nach Einzelbildfortschaltung (in Millisekunden)
} codes[] =
{
  { Panasonic_FS1,
    "DDaaabaaaaaaaaabaaabaaaaabaaababababaaabaaababaw"  // play
    "DDaaabaaaaaaaaababaaaaaaabaaababababaaaaabababa",  // & pause
    "DDaaabaaaaaaaaaaababaaaaabaaabababababaaaaababaw"  // 2x still adv.
    "DDaaabaaaaaaaaaaababaaaaabaaabababababaaaaababa",  
    "DDaaabaaaaaaaaaaaaaaaaaaabaaabababababababababa",  // stop
    750
  },
  { Samsung_SV300W,
    "BB13111311111111111311131111111111131111131311111111131311111313131w"  // play
    "BB13111311111111111311131111111111111313111311111113111113111313131",  // & pause
    "BB13111311111111111311131111111111111313111311111113111113111313131",  // pause = still adv.
    "BB13111311111111111311131111111111131113111311111111131113111313131",  // stop
    750
  },
  { Pioneer_CLD950,
    "XYssssssslssslssslslslslssslssslssslslslssslssssssssssssslssslslslsw"  // play
    "XYssssssslssslssslslslslssslssslssssssssslslssssssslslslssssslslslsL"  // pause
    "XYssssssslssslssslslslslssslssslssssssssssslssslssslslslslssslssslsW", // Step <

    "XYssssssslssslssslslslslssslssslssssssslssslssslssslslssslssslsssls",  // Step >
    "XYssssssslssslssslslslslssslssslssssslslssslssssssslssssslssslslsls",  // Stop
    250,  // bei 1/4-Aufloesung geht auch noch 125
  },
  { Sony_VCR,
    "S222422242422222424222422w"  // play
    "S242222242422222424222422" , // pause

    "S242222222424222424222422" , // Step >|
    "S222222242422222424222422" , // Stop
    750
  },
  { IRCode_Invalid, NULL, NULL, NULL }
};

static double len[256];

Remote_IR::Remote_IR(enum IRCode code)
{
  len['1'] = 13.0/22050.0;
  len['3'] = 38.0/22050.0;
  len['A'] =200.0/22050.0;
  len['B'] =100.0/22050.0;
  len['a'] = 20.0/22050.0;
  len['b'] = 60.0/22050.0;
  len['D'] = 80.0/22050.0;

  // Laserdisc
  len['X']= 375.0/44100.0;
  len['Y']= 190.0/44100.0;
  len['s']=  25.0/44100.0;
  len['l']=  70.0/44100.0;

  // Sony-VCR
  len['2']=  28.0/44100.0;
  len['4']=  53.0/44100.0;
  len['S']= 105.0/44100.0;

  // misc.
  len['w'] = 0.2;
  len['W'] = 1.0;
  len['L'] = 20.0;

  d_start=NULL;
  for (int i=0;codes[i].code != IRCode_Invalid;i++)
    if (codes[i].code == code)
      {
        d_start = codes[i].seq_start;
        d_step  = codes[i].seq_step;
        d_end   = codes[i].seq_end;
        d_waittime = codes[i].waittime;
        break;
      }
  assert(d_start!=NULL);

  cout << "------------------------------------------> REMOTE INIT\n";

  d_fd_ser = open("/dev/ttyS1",O_RDWR);
  if (d_fd_ser==-1) { perror("Err 0:"); exit(5); }

  struct termios tio;
  int i;
  if ((i=tcgetattr(d_fd_ser,&tio)))
    { perror("Err 1:"); exit(5); }
  if (cfsetospeed(&tio,                B38400))
    { perror("Err 2:"); exit(5); }
  tio.c_cflag &= ~CRTSCTS;
  if (tcsetattr(d_fd_ser,TCSANOW,&tio))
    { perror("Err 3:"); exit(5); }

  cout << "------------------------------------------> REMOTE INIT 2\n";

  Send(d_start);
  sleep(10);
  cout << "------------------------------------------> REMOTE INIT 3\n";

  if (gettimeofday(&d_sendtv,NULL)<0)
    { perror("gettimeofday failed: "); exit(10); }
}

Remote_IR::~Remote_IR()
{
  Send(d_end);
  close(d_fd_ser);
}

void Remote_IR::NextPicture()
{
  Send(d_step);

  if (gettimeofday(&d_sendtv,NULL)<0)
    { perror("gettimeofday failed: "); exit(10); }
}

void Remote_IR::WaitForNextPicture() const
{
  struct timeval tv;
  if (gettimeofday(&tv,NULL)<0)
    { perror("gettimeofday failed: "); exit(10); }

  uint64 diff_ms;
  diff_ms = (tv.tv_sec-d_sendtv.tv_sec)*1000 +
            (tv.tv_usec-d_sendtv.tv_usec)/1000;

  if (d_waittime > diff_ms)
    {
      diff_ms = d_waittime-diff_ms;
      //cout << "wait " << diff_ms << "ms\n";

      struct timespec ts;
      ts.tv_sec=diff_ms/1000;
      ts.tv_nsec=(diff_ms%1000)*1000000;
      nanosleep(&ts,NULL);
    }
}


void Remote_IR::SendPulse(int onoff) const
{
  static int state=0;
  static int sendbuffer=0;

  if (onoff)
    {
      switch (state)
	{
	case 0: sendbuffer |= 0x00; break;
	case 1: sendbuffer |= 0x04; break;
	case 2: sendbuffer |= 0x20; break;
	}
    }
  else
    {
      switch (state)
	{
	case 0: sendbuffer |= 0x01; break;
	case 1: sendbuffer |= 0x0C; break;
	case 2: sendbuffer |= 0x60; break;
	}
    }

  state++;
  if (state==3)
    {
      unsigned char n=255-sendbuffer;
      write(d_fd_ser,&n,1);
      state=0;
      sendbuffer=0;
    }
}

void Remote_IR::Send(const char* seq) const
{
  int onflag = 1;
	  
  for (int i=0;seq[i];i++)
    {
      char curr_char = seq[i];
      int length = (int)(len[curr_char]/bitlength+0.5);
	      
      for (int j=0;j<length;j++)
        {
          SendPulse(onflag);
        }
      onflag=1-onflag;
    }

  SendPulse(0);
  SendPulse(0);
  SendPulse(0);
}
